=begin

Copyright 2011, Andreas Eisenbarth
All Rights Reserved

Permission to use, copy, modify, and distribute this software for 
any purpose and without fee is hereby granted, provided that the above
copyright notice appear in all copies.

THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

Name:         TextureResizer.rb
Author:       Andreas Eisenbarth
Usage:        menu Plugins > 'Resize Textures'
Description:  This Plugin allows to downsample all textures in the model by 
              percentage or depending on texture size in model.
              This version now also considers the texture's dimensions on individual 
              faces (whereas the previous version took the dimensions as set in 
              the materials browser) and in transformed/scaled and nested components. 
              If two faces have different texture dimensions, the dimensions will
              be averaged (weighted by face area).
              Textures can also be optimized for a specific view if the plugin Goldilocks is installed.
Compatibility: Windows and Wine use supplied portable version of ImageMagick (Linux/Wine also native ImageMagick).
              OS X uses Core Image (sips).
Version:      1.2.1
Date:         01.11.2011
Credits:      Adam Billyard for some lines of Goldilocks, Copyright (c) 2009 Billyard Enterprises
              http://forums.sketchucation.com/viewtopic.php?t=22658
=end
require 'sketchup.rb'



module TextureResizer

@dir = File.dirname(__FILE__) # this plugin's folder
@temp = File.join( File.expand_path( ENV['TMPDIR'] || ENV['TMP'] || ENV['TEMP'] ), "skp_tr")
@platform = RUBY_PLATFORM.include?('darwin')? "mac" : (RUBY_PLATFORM.include?('mswin')? (!ENV["WINEPREFIX"].nil?? "wine" : "mswin") : "other")

# location of portable ImageMagick
@im_win = File.join(@dir,"IM_win/convert.exe")

# default configuration
@default_options = {
  # method of resizing; "percent" or depending on real-world "resolution"
  "method" => "percent",
  "percentage" => 50,
  "resolution" => 20,
  # whether to consider the uvq coordinates of each individual face (slow) or the dimensions in material browser (fast)
  "resolution-by-faces" => true,
  # whether to limit the possible texture dimensions to a maximum or minimum
  "limit" => false,
  "max" => 1024,
  "min" => 20,
  # whether to increase texture sizes if method is "resolution"
  "only-downsample" => true,
  # percentage that the new texture dimensions must differ from the old dimensions
  "not-modify-threshold" => 2,
  # whether to apply a specific level of jpeg compression, otherwise take it from the source file 
  "compression" => false,
  "jpeg-quality" => "75",
  # resolution to use if the plugin Goldilocks is installed
  "goldilocks-resolution" => 4
}



##
# run plugin from UI
#
def self.run
  @model = Sketchup.active_model
  @sel = @model.selection
  @goldilocks = (begin;Goldilocks.respond_to?("methods");true;rescue NameError;false;end)
  return UI.messagebox("The model does not contain any materials.", MB_OK) if @model.materials.length==0
  self.options_dialog
end #def run



##
# open the options dialog
#
def self.options_dialog
  optdlg_title = (@sel.empty?)? "Resize textures in model" : "Resize textures in selection"
  @optdlg = UI::WebDialog.new(optdlg_title, false, false, 300, 300, 475, 100, true)
  @optdlg.set_size(475, 100) # default size, will be adjusted later
  @optdlg.set_position(300, 300)
  html_path = File.join(@dir, "TextureResizerOptions.htm")
  @optdlg.set_file(html_path)
  # callbacks
  # init: pass the default options to the form
  @optdlg.add_action_callback("init") {|@optdlg, param|
    @optdlg.execute_script("document.getElementsByTagName('body')[0].style.background='"+@optdlg.get_default_dialog_color+"'") if @platform != "mac"
    @optdlg.execute_script("start(" + @default_options.inspect.gsub(/=>/,":") + ")")
    @optdlg.execute_script("s=document.getElementById('style0'); s.parentNode.removeChild(s)") if @goldilocks == true
  }
  # receives the document's inner dimensions in order to obtain the window border size
  @optdlg.add_action_callback("windowInner") {|@optdlg, params|
    params = eval(params)
    @windowBorderWidth = (475-params[0].to_f)/2
    @windowTitlebarHeight = 100-params[1].to_f
    @optdlg.execute_script("resizeHeight()")
  }
  # receives the html's dimensions in order set a specific webdialog size
  @optdlg.add_action_callback("resizeHeight") {|@optdlg, params|
    params = eval(params)
    width = 2*@windowBorderWidth+params[0].to_f; width = 300 if width<300
    height = @windowTitlebarHeight+params[1].to_f;
    @optdlg.set_size(width, height)
  }
  # OK button
  @optdlg.add_action_callback("receiveOptions") { |@optdlg, param|
    options = eval(@optdlg.get_element_value("results"))
    @default_options = @default_options.merge(options)
    @optdlg.execute_script("startProgress()")
    UI.start_timer(0.1, false){self.process(options)} # allow the webdialog to update/redraw
  }
  # Cancel button
  @optdlg.add_action_callback("cancel") { |@optdlg, param| @optdlg.close}
  # for debugging
  @optdlg.add_action_callback("puts") { |@optdlg, param|
    puts(param)
  }
  # show the webdialog
  @optdlg.show_modal
end #def options_dialog



##
# main process
#
# Can be invoked directly from ruby without UI
# optional parameters: 
#   - options hash (missing options will be taken from default options)
#   - entity collection/ array of entities/ single face, otherwise selected entities or whole model
# Loops over all entities and collects materials from entities
# Loops over collected materials, exports and create ImageMagick conversion command
# Executes all commands
# Re-imports all textures
#
def self.process(options={}, entities=((Sketchup.active_model.selection.empty?)? Sketchup.active_model.entities : Sketchup.active_model.selection) )
  @model = Sketchup.active_model
  @sel = @model.selection
  @goldilocks = (begin;Goldilocks.respond_to?("methods");true;rescue NameError;false;end)

  begin
    (Sketchup.version_number<7000000)? @model.start_operation("Resize Textures") : @model.start_operation("Resize Textures", true)
    options = @default_options.merge(options)
    entities = [entities] if entities.class==Sketchup::Face
    @materials = {}
    @instances = {} # remember each processed definition and of one instance transformation.to_a

    # (shortcut for all materials in model & percentage)
    if entities.class==Sketchup::Entities && (options["method"]=="percent" || options["resolution-by-faces"]==false || @goldilocks==true)
      @model.materials.find_all{|x| x.materialType > 0}.each{|x| @materials[x] = {}}
    # loop over all entities and collect materials in @materials hash (normal)
    else
      entities.each{|x|
        self.analyse_entity(x, options)
      }
    end

    # error if materials have no textures
    messagebox_text = "The #{(@sel.empty?)? "model" : "selection"} does not contain any textures."
    return (@optdlg.close; @model.abort_operation; UI.messagebox(messagebox_text, MB_OK)) if @materials.empty?

    # analyse model with Goldilocks, following lines based on Goldilocks, Copyright (c) 2009 Billyard Enterprises
    if @goldilocks == true && options["method"] == "goldilocks"
      width = @model.active_view.vpwidth
      height = @model.active_view.vpheight
      unitvp = Math::sqrt(width * height)
      stats = Hash.new
      w2t = Hash.new
      @model.materials.each {|mat|
		    w2t[mat] = Math::sqrt(mat.texture.image_width * mat.texture.image_height) / Math::sqrt(mat.texture.width * mat.texture.height) if mat.texture
      }
      Goldilocks.walkFaces(entities, Geom::Transformation.new, nil, stats) {|face, xform, zdist, mat|
	      tsize = Math::sqrt(face.area) * w2t[mat]
	      psize = Math::sqrt(face.area) * (unitvp / zdist)
	      tsize / psize
      }
      @materials.each{|mat, value| value["ratio"] = stats[mat] }
    end #if @goldilocks

    # export each material and created ImageMagick conversion command
    @cmds = [] # commands for batch conversion
    Dir.mkdir(@temp) if !File.exists?(@temp)
    @materials.each_key{|m|
      self.prepare_conversion(m, options)
    }

    # run batch conversion
    self.run_shell_command(@cmds){
      # re-import texture
      @materials.each{|m, value|
        self.replace_texture(m, value["path"])
      }
      @optdlg.execute_script("endProgress()")
      # clean up temporary files
      Dir.foreach(@temp){|f|
        next if f=='.' or f=='..'
        File.delete(File.join(@temp, f))
      } if File.exists?(@temp) && Dir.entries(@temp)[2]
      Dir.delete(@temp) if File.exists?(@temp)
      @optdlg.close if @optdlg!=nil
      @model.commit_operation
    }
  rescue Exception => e
    # clean up temporary files
    Dir.foreach(@temp){|f|
      next if f=='.' or f=='..'
      File.delete(File.join(@temp, f))
    } if File.exists?(@temp) && Dir.entries(@temp)[2]
    Dir.delete(@temp) if File.exists?(@temp)
    @optdlg.close if @optdlg!=nil
    @model.abort_operation
    puts("Error #<#{e.class.name}:#{e.message}.>")
    puts(e.backtrace) if $VERBOSE
  end #begin/rescue
end #def process



##########



##
# add materials of entity to hash, recursive
#
def self.analyse_entity(x, options, scale=1)
  # face with front material and/or back material
  if x.class == Sketchup::Face
    [x.material, x.back_material].each_with_index{|m, i|
      next if m==nil || m.materialType==0
      # shortcut for percent: no need to calculate texture scale on each face
      next (@materials[m] = {}) if  options["method"]=="percent" || options["resolution-by-faces"]==false || @goldilocks==true
      # add it to the hash if not yet included
      side = (i==0)? true : false
      if !@materials.include?(m)
        @materials[m] = {}
        area = x.area
        dim = self.texture_dimensions(x, side)
        @materials[m]["dim"] = []
        @materials[m]["dim"][0] = dim[0] * scale * area
        @materials[m]["dim"][1] = dim[1] * scale * area
        @materials[m]["area"] = area
      else
        area = x.area
        dim = self.texture_dimensions(x, side)
        @materials[m]["dim"][0] += dim[0] * scale * area
        @materials[m]["dim"][1] += dim[1] * scale * area
        @materials[m]["area"] += area
      end
    }
  # instance: loop over its definition
  elsif x.class == Sketchup::Group || x.class == Sketchup::ComponentInstance
    definition = (x.class == Sketchup::Group)? x.entities.parent : x.definition
    # shortcut: if instances have all the same transformation, process only one instance of each definition
    t = x.transformation.to_a
    return if @instances.include?(definition) && t == @instances[definition]
    @instances[definition] = t if !@instances.include?(definition)
    # average scale of the instance:
    t_scale = (Geom::Vector3d.new(t[0],t[4],t[8]).length+Geom::Vector3d.new(t[1],t[5],t[9]).length+Geom::Vector3d.new(t[2],t[6],t[10]).length)/3.0
    definition.entities.each{|y|
      self.analyse_entity(y, options, t_scale*scale)
    }
  end
end #def analyse_entity



##
# get the dimensions of a texture on a face 
# (best no distortion, fallback for projective distortion)
#
def self.texture_dimensions(face, front=true)
  w = (front)? face.material.texture.width : face.back_material.texture.width
  h = (front)? face.material.texture.height : face.back_material.texture.height
  return [w, h] if texture_distorted?(face, front)
  # faster than doing a real projective mapping (xyz <-> uvq)
  tw = Sketchup.create_texture_writer
  uv_helper = face.get_UVHelper(true, true, tw)
  # three points in object space
  v0 = face.vertices[0].position
  v1 = face.vertices[1].position
  vec0 = v0.vector_to(v1)
  vec1 = vec0 * face.normal
  v2 = v0 + vec1
  # in texture space
  if front==true
    uv0 = uv_helper.get_front_UVQ(v0)
    uv1 = uv_helper.get_front_UVQ(v1)
    uv2 = uv_helper.get_front_UVQ(v2)
  else
    uv0 = uv_helper.get_back_UVQ(v0)
    uv1 = uv_helper.get_back_UVQ(v1)
    uv2 = uv_helper.get_back_UVQ(v2)
  end
  uv0 = uv0.to_a.collect!{|x| x /= uv0[2]}
  uv1 = uv1.to_a.collect!{|x| x /= uv1[2]}
  uv2 = uv2.to_a.collect!{|x| x /= uv2[2]}
  # compare the object space lengths to texture space lengths
  uvec0 = uv0.vector_to(uv1)
  uvec1 = uv0.vector_to(uv2)
  return [w, h] if uvec0.length==0 || uvec1.length==0
  d0 = vec0.length/uvec0.length
  d1 = vec1.length/uvec1.length
  # we don't know which is width/height, 
  # so we take the maximum because ImageMagick will fit the image in
  maxdimension = [d0, d1].max
  if w < h
    w = maxdimension*w/h.to_f
    h = maxdimension
  else
    h = maxdimension*h/w.to_f
    w = maxdimension
  end
  return [w, h]
end



## 
# test if a face has a projective (not affine) distortion
#   based on Tomasz Marek's su2kt.rb: SU2KT.texture_distorted
#   SketchUp's texturewriter writes a unique file for projective distortions.
#   Thus a projective texture gets a different handle, 
#   while two affine textures get the same handle.
#
def self.texture_distorted?(face, front=true)
  return false if face.class!=Sketchup::Face
	distorted=false
	temp_tw = Sketchup.create_texture_writer
  ents = @model.active_entities
	@model.start_operation "Group"
	group = ents.add_group
	group.material = (front==true)? face.material : face.back_material
	group_handle = temp_tw.load(group)
	temp_handle = temp_tw.load(face, front)
	ents.erase_entities(group)
	Sketchup.undo
	distorted = true if temp_handle != group_handle
	temp_tw = nil
	return distorted
end



##
# - export a given <material> to a file
# - generate the conversion command for ImageMagick from the <options> hash
#
def self.prepare_conversion(material, options)
  texturepath = to_filename(@temp, File.basename(material.texture.filename))
  pw = material.texture.image_width # pixel width
  ph = material.texture.image_height # pixel height
  rw = material.texture.width.to_m # real-world width
  rh = material.texture.height.to_m # real-world height
  if @materials && @materials[material]["dim"]
    rw = @materials[material]["dim"][0].to_m/@materials[material]["area"]
    rh = @materials[material]["dim"][1].to_m/@materials[material]["area"]
  end
  percentage = options["percentage"].to_f
  resolution = options["resolution"].to_f
  max = options["max"].to_i
  min = options["min"].to_i

  # calculate new dimensions
  if options["method"] == "percent"
    nw = (percentage/100*pw).to_i # new pixel width
    nh = (percentage/100*ph).to_i # new pixel height
  elsif options["method"] == "resolution"
    nw = (resolution.to_f*rw).to_i
    nh = (resolution.to_f*rh).to_i
  elsif @goldilocks == true && options["method"] == "goldilocks"
    ratio = options["goldilocks-resolution"].to_f/@materials[material]["ratio"].to_f
    return (@materials.delete(material);) if ratio > 1
    nw = (ratio*pw).to_i
    nh = (ratio*pw).to_i
  end

  # avoid zero pixel dimensions or color artefacts at too small sizes
  nw = 2 if nw < 2
  nh = 2 if nh < 2

  # optionally skip texture if new dimensions would be bigger than old dimensions
  return (@materials.delete(material); puts "skip "+material.name+": no need to downsample") if options["only-downsample"] == true && options["method"] == "resolution" && (nw>pw+1 || nh>ph+1)

  # compose command
  if @platform == "mac" # use sips
    if options["limit"] == true && (nw|nh) > max
      resize_option = max.to_s
    elsif options["limit"] == true && (nw|nh) < min
      resize_option = ( [nw,nh].max/[nw/nh].min.to_f*min ).to_i.to_s
    else
      resize_option = [nw,nh].max.to_s
    end
    # set quality for jpg or jpeg textures
    resize_option += " -s formatOptions " + options["jpeg-quality"].to_s if options["compression"] == true && texturepath[/\.(?:jpg|jpeg)$/]!=nil
    # compose command for sips
    # example: sips input.jpg --resampleHeightWidthMax 256 -s formatOptions 75
    image = '"' + texturepath + '"'
    cmd = 'sips ' + image + ' -Z ' + resize_option

  else # use ImageMagick
    # set limit
    if options["limit"] == true && (nw|nh) > max
      resize_option = max.to_s + "x" + max.to_s
    elsif options["limit"] == true && (nw|nh) < min
      resize_option = min.to_s + "x" + min.to_s + "^^" # = escaped ^
    else
      resize_option = nw.to_s + "x" + nh.to_s
    end
    # skip texture if the new texture dimensions differ less than <threshold> % from the old dimensions
    return (@materials.delete(material); puts "skip "+material.name+": new dimensions almost identical to old dimensions") if (100*(1-nw/pw.to_f).abs < options["not-modify-threshold"] || 100*(1-nh/ph.to_f).abs < options["not-modify-threshold"]) && options["compression"] != true
    # set quality for jpg or jpeg textures
    resize_option += " -quality " + options["jpeg-quality"].to_s if options["compression"] == true && texturepath[/\.(?:jpg|jpeg)$/]!=nil
    # compose command for ImageMagick
    # example: convert.exe input.jpg -resize 50% output.jpg)
    # resize_option: <width>x<height>{%}{@}{!}{<}{>}
    if @platform == "wine" && File.exists?("Z:/usr/bin/convert")
      image = %["`winepath -u '#{texturepath}'`"]
      cmd = 'convert ' + image + ' -resize ' + resize_option + ' ' + image
    elsif @platform == "mswin" || @platform == "wine"
      executable = '"' + @im_win.gsub(/\//,'//').gsub(/\//,'\\') + '"'
      image = '"' + texturepath.gsub(/\//,'//').gsub(/\//,'\\') + '"'
      cmd = executable + ' ' + image + ' -resize ' + resize_option + ' ' + image
    end
  end

  # export texture path
  tw = Sketchup.create_texture_writer
   # create temporary group to export texture image without uv-mapping
  tmp_group = @model.entities.add_group
  tmp_group.material = material
  tw.load(tmp_group)
  success = tw.write(tmp_group, texturepath, true)
  return (@materials.delete(material); puts material.name+": could not write file "+texturepath) if success!=0
  @materials[material]["path"] = texturepath
  @cmds.push(cmd)
end #def prepare_conversion



##
# Execute a command / a list of commands <cmd> and continues with a callback <block>
#
# modified, based on Fredo6's script http://forums.sketchucation.com/viewtopic.php?f=180&t=36973
# - on Windows: launches a vbs script to avoid a black console window
# - launches a batch file asynchronously in order not to run in SketchUp's process
# - continues optionally with a callback, after the batch script is finished
#
def self.run_shell_command(cmd, &block)
  cmd = [cmd] if cmd.class != Array
  # Windows
  if @platform == "mswin"
    # create a batch script with the commands
    fbat = File.join(@temp, "temporary.bat")
    File.open(fbat, "w"){|f|
      f.puts "@echo off"
      cmd.each{|c| f.puts(c) }
      f.puts %[echo "finished" > "#{@temp.gsub(/\//,'//').gsub(/\//,'\\')}\\finished.txt"]
    }
    # create a Visual Basic file to launch the batch script without black console window
    fvbs = File.join(@temp, "temporary.vbs")
    File.open(fvbs, "w"){|f|
      f.puts %[Set WshShell = CreateObject("WScript.Shell")]
      f.puts %[WshShell.Run """#{fbat}""", 0]
      # a third argument bWaitOnReturn=true would wait for the script to finish (synchronously),
      # but uses SketchUp's process and is 20 times slower.
      # f.puts %[WshShell.Run """#{fbat}""", 0, true]
      f.puts %[WScript.Quit]
    }
    system(%[wscript #{fvbs}])

  # Mac OS X
  elsif @platform == "mac"
    fsh = File.join(@temp, "temporary.sh")
    File.open(fsh, "w"){|f|
      f.puts %[ #!/bin/sh]
      cmd.each{|c| f.print(c+";") }
      f.puts %[touch "#{@temp}/finished.txt";]
    }
    system(%[bash #{fsh}])

  # other
  elsif @platform == "wine"
    # check for native ImageMagick installation
    if File.exists?("Z:/bin/bash") && File.exists?("Z:/usr/bin/convert")
      fsh = File.join(@temp, "temporary.sh")
      File.open(fsh, "w"){|f|
        f.puts %[ #!/bin/sh]
        cmd.each{|c| f.print(c+";") }
        f.puts %[touch "`winepath -u '#{@temp}'`/finished.txt";]
      }
      begin
        system(%[Z:\\bin\\bash "#{ENV["WINEPREFIX"]}/dosdevices/#{fsh.sub(/^(\w)(?=\:)/,@dir[0,1].downcase)}"])
      rescue
      end
    # otherwise use portable Windows ImageMagick as distributed with this plugin
    #   no WScript necessary because Wine does not show black console window
    else
      # create a batch script with the commands
      fbat = File.join(@temp, "temporary.bat")
      File.open(fbat, "w"){|f|
        f.puts "@echo off"
        cmd.each{|c| f.puts(c) }
        f.puts %[echo "finished" > "#{@temp.gsub(/\//,'//').gsub(/\//,'\\')}\\finished.txt"]
      }
      system(%[#{fbat}])
    end
  end
  # since the batch script runs asynchronously, we need to know when it is finished
  # the batch script creates a "finished.txt" file, and we continue only when it is created
  self.file_observer(File.join(@temp,"finished.txt")){
    block.call
  }
end #def run_shell_command



##
# observe when a file exists and execute block
#
def self.file_observer(file, &block)
  if !File.exists?(file)
    UI.start_timer(0.5){ self.file_observer(file, &block) }
  else
    File.delete(file)
    block.call
  end
end #def file_observer



##
# re-import a texture from texturepath and fix its dimensions and color
#
def self.replace_texture(material, texturepath)
  return (puts("file "+texturepath+" not found")) if !File.exists?(texturepath)
  rw = material.texture.width # real-world width
  rh = material.texture.height # real-world height
  color = material.color
  # re-import texture
  material.texture = @materials[material]["path"]
  # fix texture size (SketchUp changes it sometimes after replacing the image file!)
  material.texture.size = [rw,rh]
  material.color = color
end #def reimport_texture



##
# Convert a random string into a proper, unique filename
# arguments:
#   dir: directory where the file should go
#   string: any user input
#   ext: force a specific extension if the string doesn't contain one
# returns: a complete file path
# 
def self.to_filename(dir, string, ext=nil)
  # separate extension from basename
  ext = string[/\.\w{1,3}$/].to_s if ext.nil?
  string.sub!(/#{ext}$/,"")
  # clean up basename
  string = string[/.{1,30}/].gsub(/\n|\r/,"").gsub(/[^0-9a-zA-Z\-\_\.\(\)\#]+/,"_").gsub(/^_+|_+$/,"").to_s
  string = "file" if string.empty?
  base = File.join(dir, string)
  # detect collision of filenames and return alternative filename (numbered)
  base_orig = base
  i = 0
  while File.exists?(base + ext)
    base = base_orig + i.to_s
    i += 1
  end
  return base + ext
end #def to_filename



end #module TextureResizer



if( not file_loaded?(File.basename(__FILE__)) )
  cmd_resize = UI::Command.new("Resize Textures"){TextureResizer.run()}
  UI.menu("Plugins").add_item(cmd_resize)
end
file_loaded(File.basename(__FILE__))
